import torch
import torch_optimizer as optim
from finol.model_layer.DNN import *
from finol.model_layer.RNN import *
from finol.model_layer.LSTM import *
from finol.model_layer.CNN import *
from finol.model_layer.Transformer import *
from finol.model_layer.LSRE_CAAN import *
from finol.config import *

model_dict = {
    'DNN': DNN,
    'RNN': RNN,
    'LSTM': LSTM,
    'CNN': CNN,
    'Transformer': Transformer,
    'LSRE-CAAN': LSRE_CAAN
}


def select_model(load_dataset_output):
    model_cls = model_dict.get(MODEL_NAME, None)
    if model_cls is None:
        raise ValueError(f"Invalid model strategy: {MODEL_NAME}")

    model = model_cls(
        num_assets=load_dataset_output['NUM_ASSETS'],
        num_features_augmented=load_dataset_output['NUM_FEATURES_AUGMENTED'],
        num_features_original=load_dataset_output['NUM_FEATURES_ORIGINAL'],
        window_size=load_dataset_output['WINDOW_SIZE']
    ).to(DEVICE)
    return model




import time

import torch
import numpy as np
import matplotlib.pyplot as plt
from finol.data_layer.data_loader import *
from finol.optimization_layer.model_trainer import *
from finol.evaluation_layer.metric_caculator import *
from finol.evaluation_layer.benchmark_loader import *
from finol.config import *


def evaluate_model(load_dataset_output, train_model_output):
    caculate_metric_output = caculate_metric(train_model_output, load_dataset_output)
    load_benchmark_output = load_benchmark(caculate_metric_output)

    return load_benchmark_output


if __name__ == '__main__':
    load_benchmark_output = load_benchmark()
    # load_dataset_output = load_dataset()
    # # train_model_output = train_model(load_dataset_output)
    # train_model_output = {
    #     "last_model": None,
    #     "best_model": None,
    #     "logdir": ROOT_PATH + '/evaluation_layer/logdir/2024-03-16_22-18-05'
    # }
    # evaluate_model_output = evaluate_model(load_dataset_output, train_model_output)



import time

import torch
import numpy as np
import pandas as pd
from finol.config import *

if DATASET_NAME in ["SSE", "HSI", "CMEG"]:
    day = 52
else:
    day = 252


def caculate_CW(total_return):
    # Calculate the Cumulative Wealth (CW)
    return total_return[-1]


def caculate_APY(total_return):
    # Calculate the Annual Percentage Yield (APY)
    T = len(total_return)  # Total number of trading periods
    y = T / day  # Number of years
    APY = pow(total_return[-1], 1 / y) - 1  # APY formula
    return APY


def caculate_SR(daily_return, APY):
    # Calculate the Sharpe ratio (SR)
    sigma = np.std(daily_return, ddof=1) * pow(day, 0.5)  # Standard deviation of daily returns (annualized)
    r = 0.04  # Risk-free rate of return, reference CORN in 2011 is 0.04, reference LOAD in 2019 is 0
    SR = (APY - r) / sigma  # Sharpe ratio formula
    return SR, sigma


def caculate_MDD(total_return):
    # Calculate the Maximum DrawDown (MDD)
    mddvec = []
    T = len(total_return)
    for i in range(T):
        j = i + 1
        temp = total_return[:j]
        ddval = (max(temp) - temp[i]) / max(temp)  # Drawdown value for each trading period
        mddvec.append(ddval)
    MDD = max(mddvec)  # Maximum drawdown value
    return MDD, mddvec


def caculate_ATO(NUM_PERIODS, NUM_ASSETS, model, test_loader):
    portfolio_o = np.zeros(NUM_ASSETS)
    label_list = []
    portfolio_list = []
    daily_turno_list = []
    # start_time = time.time()
    for i, data in enumerate(test_loader, 1):
        x_data, label = data
        label = label.cpu().detach().numpy()
        label_list.append(label)

        portfolio = model(x_data.float())
        portfolio = portfolio.cpu().detach().numpy()
        portfolio_list.append(portfolio)

        daily_turno = (abs(portfolio_o - portfolio).sum())
        daily_turno_list.append(daily_turno)

        daily_return = np.sum(label * portfolio) * (1 - ((0 / 2) * daily_turno))
        portfolio_o = portfolio * label / daily_return

    ATO = sum(daily_turno_list) / (2 * (NUM_PERIODS))
    return ATO


def caculate_TCW(NUM_PERIODS, NUM_ASSETS, model, test_loader):
    df = pd.DataFrame(columns=['Rate', 'TCW'])

    TRANSACTIOS_COSTS_RATE = METRIC_CONFIG.get("PRACTICAL_METRICS")["TRANSACTIOS_COSTS_RATE"]
    TRANSACTIOS_COSTS_RATE_INTERVAL = METRIC_CONFIG.get("PRACTICAL_METRICS")["TRANSACTIOS_COSTS_RATE_INTERVAL"]

    for tc_rate in np.arange(0, TRANSACTIOS_COSTS_RATE + TRANSACTIOS_COSTS_RATE_INTERVAL, TRANSACTIOS_COSTS_RATE_INTERVAL):
        portfolio_o = np.zeros(NUM_ASSETS)
        label_list = []
        portfolio_list = []
        daily_turno_list = []
        daily_return_list = []
        # start_time = time.time()
        for i, data in enumerate(test_loader, 1):
            x_data, label = data
            label = label.cpu().detach().numpy()
            label_list.append(label)

            portfolio = model(x_data.float())
            portfolio = portfolio.cpu().detach().numpy()
            portfolio_list.append(portfolio)

            daily_turno = (abs(portfolio_o - portfolio).sum())
            daily_turno_list.append(daily_turno)

            daily_return = np.sum(label * portfolio) * (1 - ((tc_rate / 2) * daily_turno))
            portfolio_o = portfolio * label / daily_return

            daily_return_list.append(daily_return)

        result = np.cumprod(daily_return_list)
        df.loc[len(df)] = [tc_rate, result[-1]]
    return df["TCW"]


def caculate_metric(train_model_output, load_dataset_output):
    model = train_model_output["best_model"]
    logdir = train_model_output["logdir"]
    if model == None:
        model = torch.load(logdir + '/best_model_'+DATASET_NAME+'.pt')

    model.eval()
    test_loader = load_dataset_output["test_loader"]
    NUM_TEST_PERIODS = load_dataset_output["NUM_TEST_PERIODS"]
    NUM_ASSETS = load_dataset_output["NUM_ASSETS"]

    portfolios = torch.zeros((NUM_TEST_PERIODS, NUM_ASSETS))
    labels = torch.zeros((NUM_TEST_PERIODS, NUM_ASSETS))

    start_time = time.time()
    for i, data in enumerate(test_loader, 0):
        x_data, label = data
        labels[i, :] = label

        portfolio = model(x_data.float())
        portfolios[i, :] = portfolio

    runtime = time.time() - start_time

    labels = labels.cpu().detach().numpy()
    portfolios = portfolios.cpu().detach().numpy()

    daily_return = np.sum(labels * portfolios, axis=1)
    total_return = np.cumprod(daily_return)

    caculate_metric_output = {}

    if METRIC_CONFIG.get('INCLUDE_PROFIT_METRICS'):
        caculate_metric_output["CW"] = caculate_CW(total_return)
        caculate_metric_output["DCW"] = total_return
        caculate_metric_output["APY"] = caculate_APY(total_return)
        caculate_metric_output["SR"], sigma = caculate_SR(daily_return, caculate_metric_output["APY"])

    if METRIC_CONFIG.get('INCLUDE_RISK_METRICS'):
        caculate_metric_output["VR"] = sigma
        caculate_metric_output["MDD"], caculate_metric_output["DDD"] = caculate_MDD(total_return)

    if METRIC_CONFIG.get("PRACTICAL_METRICS")["INCLUDE_PRACTICAL_METRICS"]:
        caculate_metric_output["ATO"] = caculate_ATO(NUM_TEST_PERIODS, NUM_ASSETS, model, test_loader)
        caculate_metric_output["TCW"] = caculate_TCW(NUM_TEST_PERIODS, NUM_ASSETS, model, test_loader)
        caculate_metric_output["RT"] = runtime

    caculate_metric_output["logdir"] = logdir
    return caculate_metric_output


import torch.nn as nn
import torch.nn.functional as F
from finol.config import *

NUM_LAYERS = MODEL_CONFIG.get("DNN")["NUM_LAYERS"]
HIDDEN_SIZE = MODEL_CONFIG.get("DNN")["HIDDEN_SIZE"]


class DNN(nn.Module):
    def __init__(
            self,
            *,
            num_assets,
            num_features_augmented,
            num_features_original,
            window_size,
            **kwargs
    ):
        super().__init__()
        self.input_size = num_features_augmented
        self.num_feats = num_features_original
        self.output_size = num_assets
        self.num_features_augmented = num_features_augmented
        self.num_features_original = num_features_original
        self.window_size = window_size

        self.layers = nn.ModuleList()
        self.layers.append(nn.Linear(self.num_features_augmented, HIDDEN_SIZE))

        for _ in range(NUM_LAYERS):
            self.layers.append(nn.Linear(HIDDEN_SIZE, HIDDEN_SIZE))

        self.layers.append(nn.Linear(HIDDEN_SIZE, 1))
        self.relu = nn.ReLU()
        self.dropout = nn.Dropout(p=DROPOUT)

    def forward(self, x):
        # batch_size, num_assets, num_features_augmented = x.shape  # n: window size; d: number of features
        # x = x.view(batch_size, num_assets, self.window_size, self.num_features_original)

        out = x
        for layer in self.layers:
            out = layer(out)
            out = self.relu(out)
            out = self.dropout(out)

        out = out.squeeze(-1)
        portfolio = F.softmax(out, dim=-1)
        return portfolio


import os
ROOT_PATH = os.path.dirname(os.path.realpath(__file__))
PARENT_PATH = os.path.dirname(ROOT_PATH)

from finol import utils
utils.download_data()
utils.check_update(GET_LATEST_FINOL=True)

import matplotlib.pyplot as plt
plt.style.use('seaborn-talk')
from rich import print

TUTORIAL_MODE = True
TUTORIAL_NAME = "TUTORIAL_4"

# Parameters related to data_layer
DATASET_NAME = "TSE"  # Available options: NYSE(O), NYSE(N), DJIA, SP500, TSE, SSE, HSI, CMEG, CRYPTO, TUTORIAL
DATASET_SPLIT_CONFIG = {
    "NYSE(O)": {
        "TRAIN_START_TIMESTAMP": "1962-07-03",
        "TRAIN_END_TIMESTAMP": "1976-01-21",  # [3390 rows x 5 columns]
        "VAL_START_TIMESTAMP": "1976-01-22",
        "VAL_END_TIMESTAMP": "1980-07-11",  # [1130 rows x 5 columns]
        "TEST_START_TIMESTAMP": "1980-07-14",
        "TEST_END_TIMESTAMP": "1984-12-31"  # [1131 rows x 5 columns]
    },
    "NYSE(N)": {
        "TRAIN_START_TIMESTAMP": "1985-01-02",
        "TRAIN_END_TIMESTAMP": "2000-04-06",  # [3858 rows x 5 columns]
        "VAL_START_TIMESTAMP": "2000-04-07",
        "VAL_END_TIMESTAMP": "2005-05-20",  # [1286 rows x 5 columns]
        "TEST_START_TIMESTAMP": "2005-05-23",
        "TEST_END_TIMESTAMP": "2010-06-30"  # [1286 rows x 5 columns]
    },
    "DJIA": {
        "TRAIN_START_TIMESTAMP": "2001-01-14",
        "TRAIN_END_TIMESTAMP": "2002-04-01",  # [300 rows x 5 columns]
        "VAL_START_TIMESTAMP": "2002-04-02",
        "VAL_END_TIMESTAMP": "2002-08-21",  # [100 rows x 5 columns]
        "TEST_START_TIMESTAMP": "2002-08-22",
        "TEST_END_TIMESTAMP": "2003-01-14"  # [100 rows x 5 columns]
    },
    "SP500": {
        "TRAIN_START_TIMESTAMP": "1998-01-02",
        "TRAIN_END_TIMESTAMP": "2001-01-12",  # [756 rows x 5 columns]
        "VAL_START_TIMESTAMP": "2001-01-16",
        "VAL_END_TIMESTAMP": "2002-01-25",  # [256 rows x 5 columns]
        "TEST_START_TIMESTAMP": "2002-01-28",
        "TEST_END_TIMESTAMP": "2003-01-31"  # [256 rows x 5 columns]
    },
    "TSE": {
        "TRAIN_START_TIMESTAMP": "1995-01-12",
        "TRAIN_END_TIMESTAMP": "1997-05-28",  # [600 rows x 5 columns] NUM OF PERIODS IN TRAINING SETS ARE UNCERATIN
        "VAL_START_TIMESTAMP": "1997-05-29",
        "VAL_END_TIMESTAMP": "1998-03-13",  # [200 rows x 5 columns]
        "TEST_START_TIMESTAMP": "1998-03-16",
        "TEST_END_TIMESTAMP": "1998-12-31"  # [201 rows x 5 columns]
    },
    "SSE": {
        "TRAIN_START_TIMESTAMP": "2010-07-05",
        "TRAIN_END_TIMESTAMP": "2018-04-09",  # [406 rows x 5 columns]
        "VAL_START_TIMESTAMP": "2018-04-16",
        "VAL_END_TIMESTAMP": "2020-11-16",  # [136 rows x 5 columns]
        "TEST_START_TIMESTAMP": "2020-11-23",
        "TEST_END_TIMESTAMP": "2023-06-26"  # [136 rows x 5 columns]
    },
    "HSI": {
        "TRAIN_START_TIMESTAMP": "2010-07-05",
        "TRAIN_END_TIMESTAMP": "2018-04-09",  # [406 rows x 5 columns]
        "VAL_START_TIMESTAMP": "2018-04-16",
        "VAL_END_TIMESTAMP": "2020-11-16",  # [136 rows x 5 columns]
        "TEST_START_TIMESTAMP": "2020-11-23",
        "TEST_END_TIMESTAMP": "2023-06-26"  # [136 rows x 5 columns]
    },
    "CMEG": {
        "TRAIN_START_TIMESTAMP": "2010-07-05",
        "TRAIN_END_TIMESTAMP": "2018-04-09",  # [406 rows x 5 columns]
        "VAL_START_TIMESTAMP": "2018-04-16",
        "VAL_END_TIMESTAMP": "2020-11-16",  # [136 rows x 5 columns]
        "TEST_START_TIMESTAMP": "2020-11-23",
        "TEST_END_TIMESTAMP": "2023-06-26"  # [136 rows x 5 columns]
    },
    "CRYPTO": {
        "TRAIN_START_TIMESTAMP": "2017-11-09",
        "TRAIN_END_TIMESTAMP": "2021-08-22",  # [1383 rows x 5 columns]
        "VAL_START_TIMESTAMP": "2021-08-23",
        "VAL_END_TIMESTAMP": "2022-11-26",  # [461 rows x 5 columns]
        "TEST_START_TIMESTAMP": "2022-11-27",
        "TEST_END_TIMESTAMP": "2024-03-01"  # [461 rows x 5 columns]
    },
}
FEATURE_ENGINEERING_CONFIG = {
    "INCLUDE_OHLCV_FEATURES": True,
    "INCLUDE_OVERLAP_FEATURES": True,
    "INCLUDE_MOMENTUM_FEATURES": True,
    "INCLUDE_VOLUME_FEATURES": True,
    "INCLUDE_CYCLE_FEATURES": True,
    "INCLUDE_PRICE_FEATURES": True,
    "INCLUDE_VOLATILITY_FEATURES": True,
    "INCLUDE_PATTERN_FEATURES": True
}
DATA_AUGMENTATION_CONFIG = {
    "WINDOW_DATA": {
        "INCLUDE_WINDOW_DATA": True,
        "WINDOW_SIZE": 10
    }
}
SCALER = "StandardScaler"  # None, StandardScaler, MinMaxScaler, MaxAbsScaler, RobustScaler
BATCH_SIZE = {
    "NYSE(O)": 128,
    "NYSE(N)": 128,
    "DJIA": 32,
    "SP500": 64,
    "TSE": 64,
    "SSE": 32,
    "HSI": 32,
    "CMEG": 32,
    "CRYPTO": 128,
}

# Parameters related to model_layer
MODEL_NAME = "DNN"
MODEL_CONFIG = {
    "DNN": {
        "NUM_LAYERS": 1,
        "HIDDEN_SIZE": 32
    },
    "RNN": {
        "NUM_LAYERS": 1,
        "HIDDEN_SIZE": 32
    },
    "LSTM": {
        "NUM_LAYERS": 1,
        "HIDDEN_SIZE": 32
    },
    "CNN": {
        "OUT_CHANNELS": 128,
        "KERNEL_SIZE": 3,
        "STRIDE": 1,
        "HIDDEN_SIZE": 32,
    },
    "Transformer": {
        "NUM_LAYERS": 1,
        "NUM_HEADS": 1,
        "HIDDEN_SIZE": 32,
    },
    "LSRE-CAAN": {
        "NUM_LATENTS": 16,
        "LATENT_DIM": 32,
        "CROSS_HEADS": 1,
        "LATENT_HEADS": 1,
        "CROSS_DIM_HEAD": 64,
        "LATENT_DIM_HEAD": 32,
        "HIDDEN_SIZE": 32,
    },
}
DROPOUT = 0.1

# Parameters related to optimization_layer
MANUAL_SEED = 42
OPTIMIZER_NAME = "Adam"
LEARNING_RATE = 1e-4
CRITERION_NAME = "LOG_SINGLE_PERIOD_WEALTH"
LAMBDA_L2 = 0.001
DEVICE = "cpu"
NUM_EPOCHES = 100

# Parameters related to evaluation_layer
MARKERS = ['o', '^', '<', '>', 's', 'p', 'h', '+', 'x', '|', '_']
MARKEVERY = 10
ALPHA = 0.5
METRIC_CONFIG = {
    "INCLUDE_PROFIT_METRICS": True,
    "INCLUDE_RISK_METRICS": True,
    "PRACTICAL_METRICS": {
        "INCLUDE_PRACTICAL_METRICS": True,
        "TRANSACTIOS_COSTS_RATE": 1/100,
        "TRANSACTIOS_COSTS_RATE_INTERVAL": 1/1000,
    }
}
BENCHMARK_BASELINE = ["Market", "Best", "UCRP", "BCRP"]
FOLLOW_THE_WINNER = ["UP", "EG", "SCRP", "PPT", "SSPO"]
FOLLOW_THE_LOSER = ["ANTI1", "ANTI2", "PAMR", "CWMR-Var", "CWMR-Stdev", "OLMAR-S", "OLMAR-E", "RMR", "RPRT"]
PATTERN_MATCHING = ["AICTR", "KTPT"]
META_LEARNING = ["SP", "ONS", "GRW", "WAAS", "CW-OGD"]

# PLOT_ALL_1 = ["EG", "PPT", "RMR", "RPRT", "Best"]
# PLOT_ALL_2 = ["AICTR", "KTPT", "SP", "CW-OGD", "Best"]

PLOT_ALL_1 = BENCHMARK_BASELINE + [MODEL_NAME]
PLOT_ALL_2 = FOLLOW_THE_WINNER + [MODEL_NAME]
PLOT_ALL_3 = FOLLOW_THE_LOSER + [MODEL_NAME]
PLOT_ALL_4 = PATTERN_MATCHING + [MODEL_NAME]
PLOT_ALL_5 = META_LEARNING + [MODEL_NAME]



################################################################################################################
############################################# FOR TUTORIAL ONLY ################################################
################################################################################################################

if TUTORIAL_MODE:
    DATASET_NAME = "NYSE(O)"
    DATASET_SPLIT_CONFIG = {
        "NYSE(O)": {
            "TRAIN_START_TIMESTAMP": "1962-07-03",
            "TRAIN_END_TIMESTAMP": "1976-01-21",  # [3390 rows x 5 columns]
            "VAL_START_TIMESTAMP": "1976-01-22",
            "VAL_END_TIMESTAMP": "1980-07-11",  # [1130 rows x 5 columns]
            "TEST_START_TIMESTAMP": "1980-07-14",
            "TEST_END_TIMESTAMP": "1984-12-31"  # [1131 rows x 5 columns]
        },
        "NYSE(N)": {
            "TRAIN_START_TIMESTAMP": "1985-01-02",
            "TRAIN_END_TIMESTAMP": "2000-04-06",  # [3858 rows x 5 columns]
            "VAL_START_TIMESTAMP": "2000-04-07",
            "VAL_END_TIMESTAMP": "2005-05-20",  # [1286 rows x 5 columns]
            "TEST_START_TIMESTAMP": "2005-05-23",
            "TEST_END_TIMESTAMP": "2010-06-30"  # [1286 rows x 5 columns]
        },
        "DJIA": {
            "TRAIN_START_TIMESTAMP": "2001-01-14",
            "TRAIN_END_TIMESTAMP": "2002-04-01",  # [300 rows x 5 columns]
            "VAL_START_TIMESTAMP": "2002-04-02",
            "VAL_END_TIMESTAMP": "2002-08-21",  # [100 rows x 5 columns]
            "TEST_START_TIMESTAMP": "2002-08-22",
            "TEST_END_TIMESTAMP": "2003-01-14"  # [100 rows x 5 columns]
        },
        "SP500": {
            "TRAIN_START_TIMESTAMP": "1998-01-02",
            "TRAIN_END_TIMESTAMP": "2001-01-12",  # [756 rows x 5 columns]
            "VAL_START_TIMESTAMP": "2001-01-16",
            "VAL_END_TIMESTAMP": "2002-01-25",  # [256 rows x 5 columns]
            "TEST_START_TIMESTAMP": "2002-01-28",
            "TEST_END_TIMESTAMP": "2003-01-31"  # [256 rows x 5 columns]
        },
        "TSE": {
            "TRAIN_START_TIMESTAMP": "1995-01-12",
            "TRAIN_END_TIMESTAMP": "1997-05-28",  # [600 rows x 5 columns] NUM OF PERIODS IN TRAINING SETS ARE UNCERATIN
            "VAL_START_TIMESTAMP": "1997-05-29",
            "VAL_END_TIMESTAMP": "1998-03-13",  # [200 rows x 5 columns]
            "TEST_START_TIMESTAMP": "1998-03-16",
            "TEST_END_TIMESTAMP": "1998-12-31"  # [201 rows x 5 columns]
        },
        "SSE": {
            "TRAIN_START_TIMESTAMP": "2010-07-05",
            "TRAIN_END_TIMESTAMP": "2018-04-09",  # [406 rows x 5 columns]
            "VAL_START_TIMESTAMP": "2018-04-16",
            "VAL_END_TIMESTAMP": "2020-11-16",  # [136 rows x 5 columns]
            "TEST_START_TIMESTAMP": "2020-11-23",
            "TEST_END_TIMESTAMP": "2023-06-26"  # [136 rows x 5 columns]
        },
        "HSI": {
            "TRAIN_START_TIMESTAMP": "2010-07-05",
            "TRAIN_END_TIMESTAMP": "2018-04-09",  # [406 rows x 5 columns]
            "VAL_START_TIMESTAMP": "2018-04-16",
            "VAL_END_TIMESTAMP": "2020-11-16",  # [136 rows x 5 columns]
            "TEST_START_TIMESTAMP": "2020-11-23",
            "TEST_END_TIMESTAMP": "2023-06-26"  # [136 rows x 5 columns]
        },
        "CMEG": {
            "TRAIN_START_TIMESTAMP": "2010-07-05",
            "TRAIN_END_TIMESTAMP": "2018-04-09",  # [406 rows x 5 columns]
            "VAL_START_TIMESTAMP": "2018-04-16",
            "VAL_END_TIMESTAMP": "2020-11-16",  # [136 rows x 5 columns]
            "TEST_START_TIMESTAMP": "2020-11-23",
            "TEST_END_TIMESTAMP": "2023-06-26"  # [136 rows x 5 columns]
        },
        "CRYPTO": {
            "TRAIN_START_TIMESTAMP": "2017-11-09",
            "TRAIN_END_TIMESTAMP": "2021-08-22",  # [1383 rows x 5 columns]
            "VAL_START_TIMESTAMP": "2021-08-23",
            "VAL_END_TIMESTAMP": "2022-11-26",  # [461 rows x 5 columns]
            "TEST_START_TIMESTAMP": "2022-11-27",
            "TEST_END_TIMESTAMP": "2024-03-01"  # [461 rows x 5 columns]
        },
    }
    FEATURE_ENGINEERING_CONFIG = {
        "INCLUDE_OHLCV_FEATURES": True,
        "INCLUDE_OVERLAP_FEATURES": True,
        "INCLUDE_MOMENTUM_FEATURES": True,
        "INCLUDE_VOLUME_FEATURES": True,
        "INCLUDE_CYCLE_FEATURES": True,
        "INCLUDE_PRICE_FEATURES": True,
        "INCLUDE_VOLATILITY_FEATURES": True,
        "INCLUDE_PATTERN_FEATURES": True
    }
    DATA_AUGMENTATION_CONFIG = {
        "WINDOW_DATA": {
            "INCLUDE_WINDOW_DATA": True,
            "WINDOW_SIZE": 10
        }
    }
    SCALER = "MinMaxScaler"  # None, StandardScaler, MinMaxScaler, MaxAbsScaler, RobustScaler
    BATCH_SIZE = {
        "NYSE(O)": 128,
        "NYSE(N)": 128,
        "DJIA": 32,
        "SP500": 64,
        "TSE": 64,
        "SSE": 32,
        "HSI": 32,
        "CMEG": 32,
        "CRYPTO": 128,
    }

    # Parameters related to model_layer
    MODEL_NAME = "LSTM"
    MODEL_CONFIG = {
        "DNN": {
            "NUM_LAYERS": 1,
            "HIDDEN_SIZE": 32
        },
        "RNN": {
            "NUM_LAYERS": 1,
            "HIDDEN_SIZE": 32
        },
        "LSTM": {
            "NUM_LAYERS": 1,
            "HIDDEN_SIZE": 32
        },
        "CNN": {
            "OUT_CHANNELS": 128,
            "KERNEL_SIZE": 3,
            "STRIDE": 1,
            "HIDDEN_SIZE": 32,
        },
        "Transformer": {
            "NUM_LAYERS": 1,
            "NUM_HEADS": 1,
            "HIDDEN_SIZE": 32,
        },
        "LSRE-CAAN": {
            "NUM_LATENTS": 16,
            "LATENT_DIM": 32,
            "CROSS_HEADS": 1,
            "LATENT_HEADS": 1,
            "CROSS_DIM_HEAD": 64,
            "LATENT_DIM_HEAD": 32,
            "HIDDEN_SIZE": 32,
        },
    }
    DROPOUT = 0.1

    # Parameters related to optimization_layer
    MANUAL_SEED = 42
    OPTIMIZER_NAME = "Adam"
    LEARNING_RATE = 1e-4
    CRITERION_NAME = "LOG_SINGLE_PERIOD_WEALTH"
    LAMBDA_L2 = 0.001
    DEVICE = "cpu"
    NUM_EPOCHES = 100

    # Parameters related to evaluation_layer
    MARKERS = ['o', '^', '<', '>', 's', 'p', 'h', '+', 'x', '|', '_']
    MARKEVERY = 10
    ALPHA = 0.5
    METRIC_CONFIG = {
        "INCLUDE_PROFIT_METRICS": True,
        "INCLUDE_RISK_METRICS": True,
        "PRACTICAL_METRICS": {
            "INCLUDE_PRACTICAL_METRICS": True,
            "TRANSACTIOS_COSTS_RATE": 1 / 100,
            "TRANSACTIOS_COSTS_RATE_INTERVAL": 1 / 1000,
        }
    }

    if TUTORIAL_NAME == "TUTORIAL_1":
        DATASET_NAME = TUTORIAL_NAME
        DATASET_SPLIT_CONFIG = {
            DATASET_NAME: {
                "TRAIN_START_TIMESTAMP": "2023-05-23",
                "TRAIN_END_TIMESTAMP": "2023-10-24",
                "VAL_START_TIMESTAMP": "2023-10-25",
                "VAL_END_TIMESTAMP": "2023-11-26",
                "TEST_START_TIMESTAMP": "2023-11-27",
                "TEST_END_TIMESTAMP": "2023-12-28"
            }
        }




from finol.data_layer.data_loader import *
from finol.optimization_layer.model_trainer import *
from finol.evaluation_layer.model_evaluator import *


load_dataset_output = load_dataset()
train_model_output = train_model(load_dataset_output)
evaluate_model_output = evaluate_model(load_dataset_output, train_model_output)


import torch
import torch.nn.functional as F
from functools import wraps
from torch import nn, einsum
from einops import rearrange, repeat
from finol.config import *

NUM_LATENTS = MODEL_CONFIG.get("LSRE-CAAN")["NUM_LATENTS"]
LATENT_DIM = MODEL_CONFIG.get("LSRE-CAAN")["LATENT_DIM"]
CROSS_HEADS = MODEL_CONFIG.get("LSRE-CAAN")["CROSS_HEADS"]
LATENT_HEADS = MODEL_CONFIG.get("LSRE-CAAN")["LATENT_HEADS"]
CROSS_DIM_HEAD = MODEL_CONFIG.get("LSRE-CAAN")["CROSS_DIM_HEAD"]
LATENT_DIM_HEAD = MODEL_CONFIG.get("LSRE-CAAN")["LATENT_DIM_HEAD"]


def exists(val):
    return val is not None


def default(val, d):
    return val if exists(val) else d


def cache_fn(f):
    cache = None

    @wraps(f)
    def cached_fn(*args, _cache=True, **kwargs):
        if not _cache:
            return f(*args, **kwargs)
        nonlocal cache
        if cache is not None:
            return cache
        cache = f(*args, **kwargs)
        return cache

    return cached_fn


class PreNorm(nn.Module):
    def __init__(self, dim, fn, context_dim=None, device=None):
        super().__init__()
        self.fn = fn.to(DEVICE)
        self.norm = nn.LayerNorm(dim).to(DEVICE)
        self.norm_context = nn.LayerNorm(context_dim).to(DEVICE) if exists(context_dim) else None

    def forward(self, x, **kwargs):
        x = self.norm(x)

        if exists(self.norm_context):
            context = kwargs['context']
            normed_context = self.norm_context(context)
            kwargs.update(context=normed_context)

        return self.fn(x, **kwargs)


class QuickGELU(nn.Module):
    def __init__(self):
        super().__init__()

    def forward(self, x):
        return x * torch.sigmoid(1.702 * x)


class FeedForward(nn.Module):
    def __init__(self, dim, device=None):
        super().__init__()
        self.net = nn.Sequential(
            nn.Linear(dim, dim).to(DEVICE),
            QuickGELU(),
        ).to(DEVICE)

    def forward(self, x):
        return self.net(x).to(DEVICE)


class Attention(nn.Module):
    def __init__(self, query_dim, context_dim=None, heads=8, dim_head=64, device=None):
        super().__init__()
        inner_dim = dim_head * heads
        context_dim = default(context_dim, query_dim)
        self.scale = dim_head ** -0.5
        self.heads = heads

        self.to_q = nn.Linear(query_dim, inner_dim, bias=False).to(DEVICE)
        self.to_kv = nn.Linear(context_dim, inner_dim * 2, bias=False).to(DEVICE)
        self.to_out = nn.Linear(inner_dim, query_dim).to(DEVICE)

    def forward(self, x, context=None, mask=None):
        h = self.heads

        q = self.to_q(x).to(DEVICE)
        context = default(context, x)  # return context if exists(context) else x
        k, v = self.to_kv(context).chunk(2, dim=-1)

        q, k, v = map(lambda t: rearrange(t, 'b n (h d) -> (b h) n d', h=h), (q, k, v))

        sim = einsum('b i d, b j d -> b i j', q, k) * self.scale

        if exists(mask):
            mask = rearrange(mask, 'b ... -> b (...)')
            max_neg_value = -torch.finfo(sim.dtype).max
            mask = repeat(mask, 'b j -> (b h) () j', h=h).to(DEVICE)
            sim.masked_fill_(~mask, max_neg_value).to(DEVICE)

        # attention, what we cannot get enough of
        attn = sim.softmax(dim=-1).to(DEVICE)

        out = einsum('b i j, b j d -> b i d', attn, v).to(DEVICE)
        out = rearrange(out, '(b h) n d -> b n (h d)', h=h).to(DEVICE)
        return self.to_out(out).to(DEVICE)


class LSRE(nn.Module):
    r"""
    This class implements the LSRE model proposed in my paper

    For more details, please refer to the papers `Online portfolio management via deep reinforcement learning with
    high-frequency data <https://www.sciencedirect.com/science/article/abs/pii/S030645732200348X>` and `Perceiver IO: A
    General Architecture for Structured Inputs & Outputs <https://arxiv.org/abs/2107.14795>`
    """

    def __init__(
            self,
            *,
            depth,
            dim,
            num_latents,
            latent_dim,
            cross_heads,
            latent_heads,
            cross_dim_head,
            latent_dim_head,
            weight_tie_layers=True,
            device,
    ):
        super().__init__()
        self.latents = nn.Parameter(torch.randn(num_latents, latent_dim)).to(DEVICE)
        self.cross_attend_blocks = nn.ModuleList([
            PreNorm(latent_dim, Attention(latent_dim, dim, heads=cross_heads, dim_head=cross_dim_head, device=DEVICE),
                    context_dim=dim, device=DEVICE).to(DEVICE),
            PreNorm(latent_dim, FeedForward(latent_dim, device=device).to(DEVICE), device=DEVICE)
        ])
        get_latent_attn = lambda: PreNorm(latent_dim,
                                          Attention(latent_dim, heads=latent_heads, dim_head=latent_dim_head,
                                                    device=DEVICE).to(DEVICE), device=DEVICE)
        get_latent_ff = lambda: PreNorm(latent_dim, FeedForward(latent_dim, device=DEVICE).to(device), device=DEVICE)
        get_latent_attn, get_latent_ff = map(cache_fn, (get_latent_attn, get_latent_ff))

        self.layers = nn.ModuleList([])
        cache_args = {'_cache': weight_tie_layers}

        for i in range(depth):
            self.layers.append(nn.ModuleList([
                get_latent_attn(**cache_args),
                get_latent_ff(**cache_args)
            ]))

    def forward(
            self,
            data,
            mask=None,
            queries=None
    ):
        # b, *_, device = *data.shape, data.device
        b, *_ = *data.shape, data.device

        # latents
        x = repeat(self.latents, 'n d -> b n d', b=b)
        cross_attn, cross_ff = self.cross_attend_blocks

        x = cross_attn(x, context=data, mask=mask) + x
        x = cross_ff(x) + x

        # layers
        for self_attn, self_ff in self.layers:
            x = self_attn(x) + x
            x = self_ff(x) + x  # x.shape = torch.Size([num_assets, num_latents, latent_dim])

        return torch.mean(x, dim=1)  # [batch_size, num_latents, latent_dim] -> [batch_size, latent_dim]


class LSRE_CAAN(nn.Module):
    r"""
    This class implements the LSRE_CAAN model

    For more details, please refer to the paper `Online portfolio management via deep reinforcement learning with
    high-frequency data <https://www.sciencedirect.com/science/article/abs/pii/S030645732200348X>`
    """

    def __init__(
            self,
            *,
            num_assets,
            num_features_augmented,
            num_features_original,
            window_size,
            **kwargs
    ):
        super().__init__()
        self.num_assets = num_assets
        self.num_features_augmented = num_features_augmented
        self.num_features_original = num_features_original
        self.window_size = window_size
        self.dim = num_features_original
        self.Prop_winners = 1
        # self.token_emb = nn.Linear(num_features_original, self.dim).to(DEVICE)
        self.pos_emb = nn.Embedding(window_size, self.dim).to(DEVICE)
        self.latent_dim = LATENT_DIM

        depth = 1
        self.lsre = LSRE(
            depth=depth,  # 1
            dim=self.dim,  # num_feats
            num_latents=NUM_LATENTS,  # 1
            # latent_dim=LATENT_DIM,  # 32
            latent_dim=self.latent_dim,  # 32
            cross_heads=CROSS_HEADS,  # 1
            latent_heads=LATENT_HEADS,  # 1
            cross_dim_head=CROSS_DIM_HEAD,  # 64
            latent_dim_head=LATENT_DIM_HEAD,  # 32
            device=DEVICE,
            **kwargs
        )

        value_dim = LATENT_DIM
        self.linear_query = torch.nn.Linear(value_dim, value_dim).to(DEVICE)
        self.linear_key = torch.nn.Linear(value_dim, value_dim).to(DEVICE)
        self.linear_value = torch.nn.Linear(value_dim, value_dim).to(DEVICE)
        self.linear_winner = torch.nn.Linear(value_dim, 1).to(DEVICE)
        self.dropout = nn.Dropout(p=DROPOUT)

        self.lsre_linear = torch.nn.Linear(LATENT_DIM, 1).to(DEVICE)

    def forward(
            self,
            x
    ):
        batch_size, num_assets, num_features_augmented = x.shape  # n: window size; d: number of features
        window_size = self.window_size
        num_features_original = self.num_features_original

        x = x.view(batch_size, num_assets, window_size, num_features_original)
        x = rearrange(x, 'b m n d -> (b m) n d')

        n, d = x.shape[1], x.shape[2]  # n: window size; d: number of features

        # LSRE
        # x = self.token_emb(x)  # optional
        pos_emb = self.pos_emb(torch.arange(n, device=DEVICE))

        pos_emb = rearrange(pos_emb, 'n d -> () n d')
        x = x + pos_emb
        stock_rep = self.lsre(x, mask=None, queries=None)  # [batch_size * num_assets, latent_dim]
        stock_rep = self.dropout(stock_rep)

        # CAAN
        x = stock_rep.view(batch_size, num_assets, self.latent_dim)

        query = self.linear_query(x)  # [batch_size, num_assets, LATENT_DIM]
        key = self.linear_key(x)  # [batch_size, num_assets, LATENT_DIM]
        value = self.linear_value(x)  # [batch_size, num_assets, LATENT_DIM]

        beta = torch.matmul(query, key.transpose(1, 2)) / torch.sqrt(torch.tensor(float(query.shape[-1])))  # [BATCH_SIZE, num_assets, num_assets]
        beta = F.softmax(beta, dim=-1).unsqueeze(-1)
        x = torch.sum(value.unsqueeze(1) * beta, dim=2)  # [batch_size, num_assets, latent_dim]

        final_scores = self.linear_winner(x).squeeze()  # [BATCH_SIZE, NUM_ASSETS]

        # Portfolio Management
        # if self.Prop_winners != 1:
        #     # Prop_winners: proportion of winners, i.e. G in Section 4.2
        #     num_winners = int(self.num_assets * self.Prop_winners)
        #     assert num_winners != 0 and num_winners <= self.num_assets
        #     rank = torch.argsort(final_scores)
        #     winners = set(rank.detach().cpu().numpy()[-num_winners:])  # <class 'set'>
        #     winners_mask = torch.Tensor([0 if i in winners else 1 for i in range(rank.shape[0])]).to(DEVICE)
        #     portfolio = F.softmax(final_scores - 1e9 * winners_mask, dim=0)
        # else:

        portfolio = F.softmax(final_scores, dim=-1)
        return portfolio



import torch.nn as nn
import torch.nn.functional as F
from einops import rearrange
from finol.config import *

NUM_LAYERS = MODEL_CONFIG.get("LSTM")["NUM_LAYERS"]
HIDDEN_SIZE = MODEL_CONFIG.get("LSTM")["HIDDEN_SIZE"]


class LSTM(nn.Module):
    def __init__(
            self,
            *,
            num_assets,
            num_features_augmented,
            num_features_original,
            window_size,
            **kwargs
    ):
        super().__init__()

        self.num_assets = num_assets
        self.num_features_augmented = num_features_augmented
        self.num_features_original = num_features_original
        self.window_size = window_size

        self.lstm = nn.LSTM(input_size=num_features_original, hidden_size=HIDDEN_SIZE, num_layers=NUM_LAYERS, batch_first=True)
        self.dropout = nn.Dropout(DROPOUT)
        self.fc = nn.Linear(HIDDEN_SIZE, 1)


    def forward(self, x):
        batch_size = x.shape[0]

        # Input Transformation
        x = x.view(batch_size, self.num_assets, self.window_size, self.num_features_original)
        x = rearrange(x, 'b m n d -> (b m) n d')

        # Temporal Representation Extraction
        out, _ = self.lstm(x)
        out = self.dropout(out)
        out = out[:, -1, :]

        # Decision Making
        out = out.view(batch_size, self.num_assets, HIDDEN_SIZE)
        out = self.fc(out).squeeze(-1)
        portfolio = F.softmax(out, dim=-1)
        return portfolio


import sys
import subprocess
import os

from finol.update import get_latest_version
from finol import __version__
# from finol.config import *


ROOT_PATH = os.path.dirname(os.path.realpath(__file__))

def check_update(GET_LATEST_FINOL):
    latest = get_latest_version()
    if __version__ == latest:
        print("The current FinOL is latest")
    else:
        print("The current FinOL is not latest, please update by `pip install --upgrade finol`")
        if GET_LATEST_FINOL:
            sys.exit()


def download_data():
    github_url = 'https://github.com/ai4finol/finol_data.git'
    local_path = ROOT_PATH + r'\data'
    subprocess.run(['git', 'clone', github_url, local_path])



from sklearn import preprocessing
from finol.config import *

scaler_dict = {
    'None': 'original data',
    'StandardScaler': preprocessing.StandardScaler,
    'MinMaxScaler': preprocessing.MinMaxScaler,
    'MaxAbsScaler': preprocessing.MaxAbsScaler,
    'RobustScaler': preprocessing.RobustScaler,
}


def select_scaler():
    scaler_cls = scaler_dict.get(SCALER, None)
    if scaler_cls is None:
        raise ValueError(f"Invalid scaler name: {SCALER}. Supported scaler are: {scaler_dict.keys()}")

    if scaler_cls == 'original data':
        return None

    scaler = scaler_cls()
    return scaler


import torch
import torch.nn as nn
import torch.nn.functional as F
from einops import rearrange
from finol.config import *

OUT_CHANNELS = MODEL_CONFIG.get("CNN")["OUT_CHANNELS"]
KERNEL_SIZE = MODEL_CONFIG.get("CNN")["KERNEL_SIZE"]
STRIDE = MODEL_CONFIG.get("CNN")["STRIDE"]
HIDDEN_SIZE = MODEL_CONFIG.get("CNN")["HIDDEN_SIZE"]


class CNN(nn.Module):
    def __init__(
            self,
            *,
            num_assets,
            num_features_augmented,
            num_features_original,
            window_size,
            **kwargs
    ):
        super(CNN, self).__init__()
        self.num_assets = num_assets
        self.num_features_augmented = num_features_augmented
        self.num_features_original = num_features_original
        self.window_size = window_size

        self.net = nn.Sequential(
            nn.Conv1d(num_features_original, OUT_CHANNELS, KERNEL_SIZE, STRIDE),
            nn.ReLU(),
            nn.AdaptiveAvgPool1d(1),
            nn.Flatten(),
            nn.Linear(OUT_CHANNELS, HIDDEN_SIZE),
            nn.ReLU(),
            nn.Linear(HIDDEN_SIZE, 1),
        )

    def forward(self, x):
        batch_size = x.shape[0]

        # # Input Transformation
        x = x.view(batch_size, self.num_assets, self.window_size, self.num_features_original)
        x = rearrange(x, 'b m n d -> (b m) n d')
        x = x.transpose(1, 2)  # [batch_size * num_assets, seq_len, num_inputs] -> [batch_size * num_assets, num_inputs, seq_len]

        # Temporal Representation Extraction
        out = self.net(x)
        out = out.view(batch_size, self.num_assets, 1).squeeze(-1)

        # Decision Making
        portfolio = F.softmax(out, dim=-1)
        return portfolio



__version__ = "0.0.6"
__author__ = "Jiahao Li"

from finol.data_layer import *
from finol.evaluation_layer import *
from finol.model_layer import *
from finol.optimization_layer import *
from .update import *
from .utils import *



import os
import time
import torch

import pandas as pd
import talib as ta
import numpy as np
from tqdm import tqdm
import matplotlib.pyplot as plt

from torch.utils.data import TensorDataset
from torch.utils.data import DataLoader

from finol.data_layer.scaler_selector import *
from finol.config import *


def data_accessing(folder_path):
    raw_files = []
    for file_name in tqdm(os.listdir(folder_path), desc="Data Loading"):
        if file_name.endswith('.xlsx') or file_name.endswith('.xls'):
            file_path = os.path.join(folder_path, file_name)
            try:
                dataframe = pd.read_excel(file_path)
                raw_files.append(dataframe)
            except Exception as e:
                print(f"An error occurred while loading file {file_path}: {str(e)}")
    return raw_files


def feature_engineering(df):
    overlap_features_df = pd.DataFrame()
    momentum_features_df = pd.DataFrame()
    volume_features_df = pd.DataFrame()
    cycle_features_df = pd.DataFrame()
    price_features_df = pd.DataFrame()
    volatility_features_df = pd.DataFrame()
    pattern_features_df = pd.DataFrame()

    if FEATURE_ENGINEERING_CONFIG.get('INCLUDE_OVERLAP_FEATURES', True):
        overlap_features = {
            'BBANDS_UPPER': ta.BBANDS(df.CLOSE)[0],  # Bollinger Bands - Upper Band
            'BBANDS_MIDDLE': ta.BBANDS(df.CLOSE)[1],  # Bollinger Bands - Middle Band
            'BBANDS_LOWER': ta.BBANDS(df.CLOSE)[2],  # Bollinger Bands - Lower Band
            'DEMA': ta.DEMA(df.CLOSE),  # Double Exponential Moving Average
            'EMA': ta.EMA(df.CLOSE),  # Exponential Moving Average
            'HT_TRENDLINE': ta.HT_TRENDLINE(df.CLOSE),  # Hilbert Transform - Instantaneous Trendline
            'KAMA': ta.KAMA(df.CLOSE),  # Kaufman Adaptive Moving Average
            'MA': ta.MA(df.CLOSE),  # Moving Average
            'MAMA': ta.MAMA(df.CLOSE)[0],  # MESA Adaptive Moving Average - MAMA
            'MAMA_FAMA': ta.MAMA(df.CLOSE)[1],  # MESA Adaptive Moving Average - FAMA
            'MAVP': ta.MAVP(df.CLOSE, df.DATE),  # Moving Average with Variable Period
            'MIDPOINT': ta.MIDPOINT(df.CLOSE),  # MidPoint over Period
            'MIDPRICE': ta.MIDPRICE(df.HIGH, df.LOW),  # Midpoint Price over Period
            'SAR': ta.SAR(df.HIGH, df.LOW),  # Parabolic SAR
            'SAREXT': ta.SAREXT(df.HIGH, df.LOW),  # Parabolic SAR - Extended
            'SMA': ta.SMA(df.CLOSE),  # Simple Moving Average
            'T3': ta.T3(df.CLOSE),  # Triple Exponential Moving Average (T3)
            'TEMA': ta.TEMA(df.CLOSE),  # Triple Exponential Moving Average
            'TRIMA': ta.TRIMA(df.CLOSE),  # Triangular Moving Average
            'WMA': ta.WMA(df.CLOSE)  # Weighted Moving Average
        }
        overlap_features_df = pd.DataFrame(overlap_features)

    if FEATURE_ENGINEERING_CONFIG.get('INCLUDE_MOMENTUM_FEATURES', True):
        momentum_features = {
            'ADX': ta.ADX(df.HIGH, df.LOW, df.CLOSE),  # Average Directional Movement Index
            'ADXR': ta.ADXR(df.HIGH, df.LOW, df.CLOSE),  # Average Directional Movement Index Rating
            'APO': ta.APO(df.CLOSE),  # Absolute Price Oscillator
            'AROON_UP': ta.AROON(df.HIGH, df.LOW)[0],  # Aroon Up
            'AROON_DOWN': ta.AROON(df.HIGH, df.LOW)[1],  # Aroon Down
            'AROONOSC': ta.AROONOSC(df.HIGH, df.LOW),  # Aroon Oscillator
            'BOP': ta.BOP(df.OPEN, df.HIGH, df.LOW, df.CLOSE),  # Balance Of Power
            'CCI': ta.CCI(df.HIGH, df.LOW, df.CLOSE),  # Commodity Channel Index
            'CMO': ta.CMO(df.CLOSE),  # Chande Momentum Oscillator
            'DX': ta.DX(df.HIGH, df.LOW, df.CLOSE),  # Directional Movement Index
            'MACD': ta.MACD(df.CLOSE)[0],  # Moving Average Convergence/Divergence
            'MACD_SIGNAL': ta.MACD(df.CLOSE)[1],  # MACD Signal Line
            'MACD_HIST': ta.MACD(df.CLOSE)[2],  # MACD Histogram
            'MACDEXT': ta.MACDEXT(df.CLOSE)[0],  # MACD with controllable MA type
            'MACDEXT_SIGNAL': ta.MACDEXT(df.CLOSE)[1],  # MACDEXT Signal Line
            'MACDEXT_HIST': ta.MACDEXT(df.CLOSE)[2],  # MACDEXT Histogram
            'MACDFIX': ta.MACDFIX(df.CLOSE)[0],  # Moving Average Convergence/Divergence Fix 12/26
            'MACDFIX_SIGNAL': ta.MACDFIX(df.CLOSE)[1],  # MACDFIX Signal Line
            'MACDFIX_HIST': ta.MACDFIX(df.CLOSE)[2],  # MACDFIX Histogram
            'MFI': ta.MFI(df.HIGH, df.LOW, df.CLOSE, df.VOLUME),  # Money Flow Index
            'MINUS_DI': ta.MINUS_DI(df.HIGH, df.LOW, df.CLOSE),  # Minus Directional Indicator
            'MINUS_DM': ta.MINUS_DM(df.HIGH, df.LOW),  # Minus Directional Movement
            'MOM': ta.MOM(df.CLOSE),  # Momentum
            'PLUS_DI': ta.PLUS_DI(df.HIGH, df.LOW, df.CLOSE),  # Plus Directional Indicator
            'PLUS_DM': ta.PLUS_DM(df.HIGH, df.LOW),  # Plus Directional Movement
            'PPO': ta.PPO(df.CLOSE),  # Percentage Price Oscillator
            'ROC': ta.ROC(df.CLOSE),  # Rate of change: ((price/prevPrice)-1)*100
            'ROCP': ta.ROCP(df.CLOSE),  # Rate of change Percentage: (price-prevPrice)/prevPrice
            'ROCR': ta.ROCR(df.CLOSE),  # Rate of change ratio: (price/prevPrice)
            'ROCR100': ta.ROCR100(df.CLOSE),  # Rate of change ratio 100 scale: (price/prevPrice)*100
            'RSI': ta.RSI(df.CLOSE),  # Relative Strength Index
            'STOCH_K': ta.STOCH(df.HIGH, df.LOW, df.CLOSE)[0],  # Stochastic %K
            'STOCH_D': ta.STOCH(df.HIGH, df.LOW, df.CLOSE)[1],  # Stochastic %D
            'STOCHF_K': ta.STOCHF(df.HIGH, df.LOW, df.CLOSE)[0],  # Stochastic Fast %K
            'STOCHF_D': ta.STOCHF(df.HIGH, df.LOW, df.CLOSE)[1],  # Stochastic Fast %D
            'STOCHRSI_K': ta.STOCHRSI(df.CLOSE)[0],  # Stochastic RSI %K
            'STOCHRSI_D': ta.STOCHRSI(df.CLOSE)[1],  # Stochastic RSI %D
            'TRIX': ta.TRIX(df.CLOSE),  # 1-day Rate-Of-Change (ROC) of a Triple Smooth EMA
            'ULTOSC': ta.ULTOSC(df.HIGH, df.LOW, df.CLOSE),  # Ultimate Oscillator
            'WILLR': ta.WILLR(df.HIGH, df.LOW, df.CLOSE)  # Williams' %R
        }
        momentum_features_df = pd.DataFrame(momentum_features)

    if FEATURE_ENGINEERING_CONFIG.get('INCLUDE_VOLUME_FEATURES', True):
        volume_features = {
            'AD': ta.AD(df.HIGH, df.LOW, df.CLOSE, df.VOLUME),  # Chaikin A/D Line
            'ADOSC': ta.ADOSC(df.HIGH, df.LOW, df.CLOSE, df.VOLUME),  # Chaikin A/D Oscillator
            'OBV': ta.OBV(df.CLOSE, df.VOLUME)  # On Balance Volume
        }
        volume_features_df = pd.DataFrame(volume_features)

    if FEATURE_ENGINEERING_CONFIG.get('INCLUDE_CYCLE_FEATURES', True):
        cycle_features = {
            'HT_DCPERIOD': ta.HT_DCPERIOD(df.CLOSE),  # Hilbert Transform - Dominant Cycle Period
            'HT_DCPHASE': ta.HT_DCPHASE(df.CLOSE),  # Hilbert Transform - Dominant Cycle Phase
            'HT_PHASOR_INPHASE': ta.HT_PHASOR(df.CLOSE)[0],  # Hilbert Transform - Phasor Components
            'HT_PHASOR_QUADRATURE': ta.HT_PHASOR(df.CLOSE)[1],  # Hilbert Transform - Phasor Components
            'HT_SINE_LEADSINE': ta.HT_SINE(df.CLOSE)[0],  # Hilbert Transform - SineWave
            'HT_SINE_SINEWAVE': ta.HT_SINE(df.CLOSE)[1],  # Hilbert Transform - SineWave
            'HT_TRENDMODE': ta.HT_TRENDMODE(df.CLOSE)  # Hilbert Transform - Trend vs Cycle Mode
        }
        cycle_features_df = pd.DataFrame(cycle_features)

    if FEATURE_ENGINEERING_CONFIG.get('INCLUDE_PRICE_FEATURES', True):
        price_features = {
            'AVGPRICE': ta.AVGPRICE(df.OPEN, df.HIGH, df.LOW, df.CLOSE),  # Average Price
            'MEDPRICE': ta.MEDPRICE(df.HIGH, df.LOW),  # Median Price
            'TYPPRICE': ta.TYPPRICE(df.HIGH, df.LOW, df.CLOSE),  # Typical Price
            'WCLPRICE': ta.WCLPRICE(df.HIGH, df.LOW, df.CLOSE)  # Weighted Close Price
        }
        price_features_df = pd.DataFrame(price_features)

    if FEATURE_ENGINEERING_CONFIG.get('INCLUDE_VOLATILITY_FEATURES', True):
        volatility_features = {
            'ATR': ta.ATR(df.HIGH, df.LOW, df.CLOSE),  # Average True Range
            'NATR': ta.NATR(df.HIGH, df.LOW, df.CLOSE),  # Normalized Average True Range
            'TRANGE': ta.TRANGE(df.HIGH, df.LOW, df.CLOSE)  # True Range
        }
        volatility_features_df = pd.DataFrame(volatility_features)

    if FEATURE_ENGINEERING_CONFIG.get('INCLUDE_PATTERN_FEATURES', True):
        pattern_features = {
            'CDL2CROWS': ta.CDL2CROWS(df.OPEN, df.HIGH, df.LOW, df.CLOSE),  # Two Crows
            'CDL3BLACKCROWS': ta.CDL3BLACKCROWS(df.OPEN, df.HIGH, df.LOW, df.CLOSE),  # Three Black Crows
            'CDL3INSIDE': ta.CDL3INSIDE(df.OPEN, df.HIGH, df.LOW, df.CLOSE),  # Three Inside Up/Down
            'CDL3LINESTRIKE': ta.CDL3LINESTRIKE(df.OPEN, df.HIGH, df.LOW, df.CLOSE),  # Three-Line Strike
            'CDL3OUTSIDE': ta.CDL3OUTSIDE(df.OPEN, df.HIGH, df.LOW, df.CLOSE),  # Three Outside Up/Down
            'CDL3STARSINSOUTH': ta.CDL3STARSINSOUTH(df.OPEN, df.HIGH, df.LOW, df.CLOSE),  # Three Stars In The South
            'CDL3WHITESOLDIERS': ta.CDL3WHITESOLDIERS(df.OPEN, df.HIGH, df.LOW, df.CLOSE),  # Three Advancing White Soldiers
            'CDLABANDONEDBABY': ta.CDLABANDONEDBABY(df.OPEN, df.HIGH, df.LOW, df.CLOSE),  # Abandoned Baby
            'CDLADVANCEBLOCK': ta.CDLADVANCEBLOCK(df.OPEN, df.HIGH, df.LOW, df.CLOSE),  # Advance Block
            'CDLBELTHOLD': ta.CDLBELTHOLD(df.OPEN, df.HIGH, df.LOW, df.CLOSE),  # Belt-Hold
            'CDLBREAKAWAY': ta.CDLBREAKAWAY(df.OPEN, df.HIGH, df.LOW, df.CLOSE),  # Breakaway
            'CDLCLOSINGMARUBOZU': ta.CDLCLOSINGMARUBOZU(df.OPEN, df.HIGH, df.LOW, df.CLOSE),  # Closing Marubozu
            'CDLCONCEALBABYSWALL': ta.CDLCONCEALBABYSWALL(df.OPEN, df.HIGH, df.LOW, df.CLOSE),  # Concealing Baby Swallow
            'CDLCOUNTERATTACK': ta.CDLCOUNTERATTACK(df.OPEN, df.HIGH, df.LOW, df.CLOSE),  # Counterattack
            'CDLDARKCLOUDCOVER': ta.CDLDARKCLOUDCOVER(df.OPEN, df.HIGH, df.LOW, df.CLOSE),  # Dark Cloud Cover
            'CDLDOJI': ta.CDLDOJI(df.OPEN, df.HIGH, df.LOW, df.CLOSE),  # Doji,
            'CDLDOJISTAR': ta.CDLDOJISTAR(df.OPEN, df.HIGH, df.LOW, df.CLOSE),  # Doji Star
            'CDLDRAGONFLYDOJI': ta.CDLDRAGONFLYDOJI(df.OPEN, df.HIGH, df.LOW, df.CLOSE),  # Dragonfly Doji
            'CDLENGULFING': ta.CDLENGULFING(df.OPEN, df.HIGH, df.LOW, df.CLOSE),  # Engulfing Pattern
            'CDLEVENINGDOJISTAR': ta.CDLEVENINGDOJISTAR(df.OPEN, df.HIGH, df.LOW, df.CLOSE),  # Evening Doji Star
            'CDLEVENINGSTAR': ta.CDLEVENINGSTAR(df.OPEN, df.HIGH, df.LOW, df.CLOSE),  # Evening Star
            'CDLGAPSIDESIDEWHITE': ta.CDLGAPSIDESIDEWHITE(df.OPEN, df.HIGH, df.LOW, df.CLOSE),  # Up/Down-Gap Side-By-Side White Lines
            'CDLGRAVESTONEDOJI': ta.CDLGRAVESTONEDOJI(df.OPEN, df.HIGH, df.LOW, df.CLOSE),  # Gravestone Doji
            'CDLHAMMER': ta.CDLHAMMER(df.OPEN, df.HIGH, df.LOW, df.CLOSE),  # Hammer
            'CDLHANGINGMAN': ta.CDLHANGINGMAN(df.OPEN, df.HIGH, df.LOW, df.CLOSE),  # Hanging Man
            'CDLHARAMI': ta.CDLHARAMI(df.OPEN, df.HIGH, df.LOW, df.CLOSE),  # Harami Pattern
            'CDLHARAMICROSS': ta.CDLHARAMICROSS(df.OPEN, df.HIGH, df.LOW, df.CLOSE),  # Harami Cross Pattern
            'CDLHIGHWAVE': ta.CDLHIGHWAVE(df.OPEN, df.HIGH, df.LOW, df.CLOSE),  # High-Wave Candle
            'CDLHIKKAKE': ta.CDLHIKKAKE(df.OPEN, df.HIGH, df.LOW, df.CLOSE),  # Hikkake Pattern
            'CDLHIKKAKEMOD': ta.CDLHIKKAKEMOD(df.OPEN, df.HIGH, df.LOW, df.CLOSE),  # Modified Hikkake Pattern
            'CDLHOMINGPIGEON': ta.CDLHOMINGPIGEON(df.OPEN, df.HIGH, df.LOW, df.CLOSE),  # Homing Pigeon
            'CDLIDENTICAL3CROWS': ta.CDLIDENTICAL3CROWS(df.OPEN, df.HIGH, df.LOW, df.CLOSE),  # Identical Three Crows
            'CDLINNECK': ta.CDLINNECK(df.OPEN, df.HIGH, df.LOW, df.CLOSE),  # In-Neck Pattern
            'CDLINVERTEDHAMMER': ta.CDLINVERTEDHAMMER(df.OPEN, df.HIGH, df.LOW, df.CLOSE),  # Inverted Hammer
            'CDLKICKING': ta.CDLKICKING(df.OPEN, df.HIGH, df.LOW, df.CLOSE),  # Kicking
            'CDLKICKINGBYLENGTH': ta.CDLKICKINGBYLENGTH(df.OPEN, df.HIGH, df.LOW, df.CLOSE),  # Kicking - Bull/Bear Determined by the Longer Marubozu
            'CDLLADDERBOTTOM': ta.CDLLADDERBOTTOM(df.OPEN, df.HIGH, df.LOW, df.CLOSE),  # Ladder Bottom
            'CDLLONGLEGGEDDOJI': ta.CDLLONGLEGGEDDOJI(df.OPEN, df.HIGH, df.LOW, df.CLOSE),  # Long Legged Doji
            'CDLLONGLINE': ta.CDLLONGLINE(df.OPEN, df.HIGH, df.LOW, df.CLOSE),  # Long Line Candle
            'CDLMARUBOZU': ta.CDLMARUBOZU(df.OPEN, df.HIGH, df.LOW, df.CLOSE),  # Marubozu
            'CDLMATCHINGLOW': ta.CDLMATCHINGLOW(df.OPEN, df.HIGH, df.LOW, df.CLOSE),  # Matching Low
            'CDLMATHOLD': ta.CDLMATHOLD(df.OPEN, df.HIGH, df.LOW, df.CLOSE),  # Mat Hold
            'CDLMORNINGDOJISTAR': ta.CDLMORNINGDOJISTAR(df.OPEN, df.HIGH, df.LOW, df.CLOSE),  # Morning Doji Star
            'CDLMORNINGSTAR': ta.CDLMORNINGSTAR(df.OPEN, df.HIGH, df.LOW, df.CLOSE),  # Morning Star
            'CDLONNECK': ta.CDLONNECK(df.OPEN, df.HIGH, df.LOW, df.CLOSE),  # On-Neck Pattern
            'CDLPIERCING': ta.CDLPIERCING(df.OPEN, df.HIGH, df.LOW, df.CLOSE),  # Piercing Pattern
            'CDLRICKSHAWMAN': ta.CDLRICKSHAWMAN(df.OPEN, df.HIGH, df.LOW, df.CLOSE),  # Rickshaw Man
            'CDLRISEFALL3METHODS': ta.CDLRISEFALL3METHODS(df.OPEN, df.HIGH, df.LOW, df.CLOSE),  # Rising/Falling Three Methods
            'CDLSEPARATINGLINES': ta.CDLSEPARATINGLINES(df.OPEN, df.HIGH, df.LOW, df.CLOSE),  # Separating Lines
            'CDLSHOOTINGSTAR': ta.CDLSHOOTINGSTAR(df.OPEN, df.HIGH, df.LOW, df.CLOSE),  # Shooting Star
            'CDLSHORTLINE': ta.CDLSHORTLINE(df.OPEN, df.HIGH, df.LOW, df.CLOSE),  # Short Line Candle
            'CDLSPINNINGTOP': ta.CDLSPINNINGTOP(df.OPEN, df.HIGH, df.LOW, df.CLOSE),  # Spinning Top
            'CDLSTALLEDPATTERN': ta.CDLSTALLEDPATTERN(df.OPEN, df.HIGH, df.LOW, df.CLOSE),  # Stalled Pattern
            'CDLSTICKSANDWICH': ta.CDLSTICKSANDWICH(df.OPEN, df.HIGH, df.LOW, df.CLOSE),  # Stick Sandwich
            'CDLTAKURI': ta.CDLTAKURI(df.OPEN, df.HIGH, df.LOW, df.CLOSE),  # Takuri (Dragonfly Doji with Very Long Lower Shadow)
            'CDLTASUKIGAP': ta.CDLTASUKIGAP(df.OPEN, df.HIGH, df.LOW, df.CLOSE),  # Tasuki Gap
            'CDLTHRUSTING': ta.CDLTHRUSTING(df.OPEN, df.HIGH, df.LOW, df.CLOSE),  # Thrusting Pattern
            'CDLTRISTAR': ta.CDLTRISTAR(df.OPEN, df.HIGH, df.LOW, df.CLOSE),  # Tristar Pattern
            'CDLUNIQUE3RIVER': ta.CDLUNIQUE3RIVER(df.OPEN, df.HIGH, df.LOW, df.CLOSE),  # Unique 3 River
            'CDLUPSIDEGAP2CROWS': ta.CDLUPSIDEGAP2CROWS(df.OPEN, df.HIGH, df.LOW, df.CLOSE),  # Upside Gap Two Crows
            'CDLXSIDEGAP3METHODS': ta.CDLXSIDEGAP3METHODS(df.OPEN, df.HIGH, df.LOW, df.CLOSE)  # Upside/Downside Gap Three Methods
        }
        pattern_features_df = pd.DataFrame(pattern_features)

    _ = pd.concat([df, overlap_features_df, momentum_features_df, volume_features_df, cycle_features_df,
                    price_features_df, volatility_features_df, pattern_features_df], axis=1)
    _.set_index('DATE', inplace=True)
    return _


def data_augmentation(df):
    if DATA_AUGMENTATION_CONFIG.get("WINDOW_DATA")["INCLUDE_WINDOW_DATA"]:
        WINDOW_SIZE = DATA_AUGMENTATION_CONFIG.get("WINDOW_DATA")["WINDOW_SIZE"]
        df = pd.concat([df] + [df.shift(i).add_prefix(f'prev_{i}_') for i in range(1, WINDOW_SIZE)], axis=1)
    else:
        WINDOW_SIZE = 1
    return df, WINDOW_SIZE


def data_cleaning(df):
    """
    Clean the DataFrame by removing rows with missing values.
    """
    return df.dropna(how='any')


def zscore_calculation(df):
    df_normalized = df.copy()
    numeric_features = df.select_dtypes(include=['int', 'float']).columns
    scaler = select_scaler()
    if scaler != None:
        zscore = scaler.fit(df_normalized[numeric_features])  # zscore is different for different assets
    else:
        zscore = None
    return zscore


def data_normalization(df, zscore):
    """
    Normalize all numeric features in DataFrame
    """
    df_normalized = df.copy()
    numeric_features = df.select_dtypes(include=['int', 'float']).columns
    if zscore != None:
        df_normalized[numeric_features] = zscore.transform(df_normalized[numeric_features])
    else:
        df_normalized[numeric_features] = df_normalized[numeric_features]
    return df_normalized


def data_splitting(df):
    """
    Split the DataFrame into train, validation, and test sets.
    """

    train_start = pd.to_datetime(DATASET_SPLIT_CONFIG.get(DATASET_NAME)["TRAIN_START_TIMESTAMP"])
    train_end = pd.to_datetime(DATASET_SPLIT_CONFIG.get(DATASET_NAME)["TRAIN_END_TIMESTAMP"])
    val_start = pd.to_datetime(DATASET_SPLIT_CONFIG.get(DATASET_NAME)["VAL_START_TIMESTAMP"])
    val_end = pd.to_datetime(DATASET_SPLIT_CONFIG.get(DATASET_NAME)["VAL_END_TIMESTAMP"])
    test_start = pd.to_datetime(DATASET_SPLIT_CONFIG.get(DATASET_NAME)["TEST_START_TIMESTAMP"])
    test_end = pd.to_datetime(DATASET_SPLIT_CONFIG.get(DATASET_NAME)["TEST_END_TIMESTAMP"])

    train = df[df.index <= train_end]
    val = df[(df.index >= val_start) & (df.index <= val_end)]
    test = df[(df.index >= test_start) & (df.index <= test_end)]
    return train, val, test


def load_dataset():
    """
    Load the raw dataset and perform some data processing operations

    Returns:
    - df (DataFrame): Processed DataFrame
    """
    ds_train = []
    ds_val = []
    ds_test = []
    label_train = []
    label_val = []
    label_test = []
    df_label_MATLAB = pd.DataFrame()

    raw_files = data_accessing(ROOT_PATH + '/data/datasets/' + DATASET_NAME)
    if len(raw_files) > 0:
        print(f"Successfully loaded {len(raw_files)} Excel file(s):")
        # feature_engineering(excel_files)
        for i, df in tqdm(enumerate(raw_files), total=len(raw_files), desc="Data Processing"):

            # print(f"Excel file {i+1}:")
            df = data_cleaning(df)

            df = feature_engineering(df)
            df = data_cleaning(df)

            df, WINDOW_SIZE = data_augmentation(df)
            df = data_cleaning(df)

            df_label = pd.DataFrame({'LABEL': df['CLOSE'].shift(-1) / df['CLOSE']})
            df_label.iloc[-1, -1] = 1

            # df_label_temp = df_label.copy()
            # test_start = pd.to_datetime(DATASET_SPLIT_CONFIG.get(DATASET_NAME)["TEST_START_TIMESTAMP"])
            # test_end = pd.to_datetime(DATASET_SPLIT_CONFIG.get(DATASET_NAME)["TEST_END_TIMESTAMP"])
            # df_label_temp = df_label_temp[(df_label_temp.index >= test_start) & (df_label_temp.index <= test_end)]
            # df_label_MATLAB = pd.concat([df_label_MATLAB, df_label_temp], axis=1)

            train, val, test = data_splitting(df)
            train_label, val_label, test_label = data_splitting(df_label)

            zscore_train = zscore_calculation(train)
            train_normalization = data_normalization(train, zscore_train)
            val_normalization = data_normalization(val, zscore_train)
            test_normalization = data_normalization(test, zscore_train)

            nan_count = df.isna().sum().sum()
            if nan_count != 0:
                print(f'nan_count: {nan_count}')
                time.sleep(1111)

            nan_count = df.isnull().sum().sum()
            if nan_count != 0:
                print(f'nan_count: {nan_count}')
                time.sleep(1111)

            train_normalization = data_cleaning(train_normalization)
            val_normalization = data_cleaning(val_normalization)
            test_normalization = data_cleaning(test_normalization)

            ds_train.append(torch.from_numpy(train_normalization.values))
            ds_val.append(torch.from_numpy(val_normalization.values))
            ds_test.append(torch.from_numpy(test_normalization.values))
            label_train.append(torch.from_numpy(train_label['LABEL'].values))
            label_val.append(torch.from_numpy(val_label['LABEL'].values))
            label_test.append(torch.from_numpy(test_label['LABEL'].values))

        ds_train = torch.stack(ds_train).permute(1, 0, 2)  # [num_assets, num_train_periods, num_feats] -> [num_train_periods, num_assets, num_feats]
        ds_val = torch.stack(ds_val).permute(1, 0, 2)
        ds_test = torch.stack(ds_test).permute(1, 0, 2)
        label_train = torch.stack(label_train).transpose(0, 1)  # [num_assets, num_train_periods] -> [num_train_periods, num_assets]
        label_val = torch.stack(label_val).transpose(0, 1)
        label_test = torch.stack(label_test).transpose(0, 1)

        train_ids = TensorDataset(ds_train, label_train)
        val_ids = TensorDataset(ds_val, label_val)
        test_ids = TensorDataset(ds_test, label_test)

        train_loader = DataLoader(train_ids, batch_size=BATCH_SIZE[DATASET_NAME], shuffle=False, drop_last=False)
        val_loader = DataLoader(val_ids, batch_size=BATCH_SIZE[DATASET_NAME], shuffle=False, drop_last=False)
        test_loader = DataLoader(test_ids, batch_size=1, shuffle=False, drop_last=False)

    else:
        print("No Excel files found.")

    # from scipy import io
    # Save the price relative tensor to a mat file with the data set name
    # io.savemat('price_relative_' + DATASET_NAME + '.mat', {'data': df_label_MATLAB.values})

    load_dataset_output = {
        "train_loader": train_loader,
        "val_loader": val_loader,
        "test_loader": test_loader,
        'NUM_TRAIN_PERIODS': ds_train.shape[0],
        'NUM_VAL_PERIODS': ds_val.shape[0],
        'NUM_TEST_PERIODS': ds_test.shape[0],
        'NUM_ASSETS': ds_train.shape[1],
        'NUM_FEATURES_AUGMENTED': ds_train.shape[2],
        'NUM_FEATURES_ORIGINAL': int(ds_train.shape[2] / (WINDOW_SIZE)),
        'WINDOW_SIZE': WINDOW_SIZE
    }
    return load_dataset_output


if __name__ == '__main__':
    # utils.check_update(GET_LATEST_FINOL)
    load_dataset_output = load_dataset()
    print(load_dataset_output)


import argparse
import json5
import matplotlib.pyplot as plt
import pandas as pd
import seaborn as sns
from finol import utils
from finol.config import *


def get_variable_name(var):
    return next(key for key, value in globals().items() if value is var)


def plot_dataframe(df, column_names, plot_type, logdir):
    plt.figure(figsize=(9, 4))

    markevery = {
        'NYSE(O)': 50,
        'NYSE(N)': 55,
        'DJIA': 3,
        'SP500': 7,
        'TSE': 6,
        'SSE': 4,
        'HSI': 4,
        'CMEG': 4,
        'CRYPTO': 13
    }.get(DATASET_NAME, 10)

    if plot_type == 'DCW':
        xlabel = 'Trading Periods'
        ylabel = 'Daily Cumulative Wealth'
    elif plot_type == 'DDD':
        xlabel = 'Trading Periods'
        ylabel = 'Daily DrawDown'
    elif plot_type == 'TCW':
        markevery = 1
        xlabel = 'Transaction Costs Rates (%)'
        ylabel = 'Costs-Adjusted Cumulative Wealth'
        plt.xticks(ticks=[0, 2, 4, 6, 8, 10], labels=[0, 0.2, 0.4, 0.6, 0.8, 1])

    df = df[column_names]
    num_columns = len(df.columns)

    colors = ['black'] * (num_columns - 1) + ['red']
    # colors = ['black'] * (num_columns)
    lines = ['-'] * (num_columns - 1) + [':']
    markers = ['o', '^', '<', '>', 's', 'p', 'h', '+', 'x', '|', '_']

    for i, column in enumerate(df.columns):
        # plt.plot(df[column], color=colors[i], marker=markers[i], markevery=markevery, markeredgecolor=colors[i],
        #          markerfacecolor='none', alpha=0.6, label=column, linewidth=2)
        plt.plot(df[column], linestyle=lines[i], color=colors[i], marker=markers[i], markevery=markevery,  alpha=0.5, label=column)

    plt.xlabel(xlabel)
    plt.ylabel(ylabel)
    plt.title(DATASET_NAME)
    plt.legend(loc='best')  # lower left
    plt.grid(True)
    # plt.yscale('log')
    plt.tight_layout()
    plt.savefig(logdir + '/' + DATASET_NAME + '_' + get_variable_name(column_names) + '_' + plot_type + '.pdf',
                format='pdf',
                dpi=300,
                bbox_inches='tight')
    plt.show()


def load_benchmark(caculate_metric_output=None):
    """
    Load baseline results from a CSV file for specified target classes.

    Returns:
    - None
    """
    if caculate_metric_output != None:
        logdir = caculate_metric_output['logdir']
    else:
        logdir = ROOT_PATH

    if METRIC_CONFIG.get('INCLUDE_PROFIT_METRICS'):
        daily_return = pd.read_excel(ROOT_PATH + '/data/benchmark_results/profit_metrics/'+DATASET_NAME+'/daily_return.xlsx')
        daily_cumulative_wealth = pd.read_excel(ROOT_PATH + '/data/benchmark_results/profit_metrics/'+DATASET_NAME+'/daily_cumulative_wealth.xlsx')
        final_profit_result = pd.read_excel(ROOT_PATH + '/data/benchmark_results/profit_metrics/'+DATASET_NAME+'/final_profit_result.xlsx')

        daily_return = daily_return.dropna(axis=1, how="any")
        daily_cumulative_wealth = daily_cumulative_wealth.dropna(axis=1, how="any")
        final_profit_result = final_profit_result.dropna(axis=1, how="any")
        if caculate_metric_output != None:
            daily_cumulative_wealth[MODEL_NAME] = caculate_metric_output["DCW"]
            final_profit_result = final_profit_result.assign(**{MODEL_NAME: 0})
            final_profit_result.loc[0, MODEL_NAME] = caculate_metric_output["CW"]
            final_profit_result.loc[1, MODEL_NAME] = caculate_metric_output["APY"]
            final_profit_result.loc[2, MODEL_NAME] = caculate_metric_output["SR"]
        print(
            final_profit_result
        )
        plot_dataframe(daily_cumulative_wealth, PLOT_ALL_1, 'DCW', logdir)
        plot_dataframe(daily_cumulative_wealth, PLOT_ALL_2, 'DCW', logdir)
        plot_dataframe(daily_cumulative_wealth, PLOT_ALL_3, 'DCW', logdir)
        plot_dataframe(daily_cumulative_wealth, PLOT_ALL_4, 'DCW', logdir)
        plot_dataframe(daily_cumulative_wealth, PLOT_ALL_5, 'DCW', logdir)

    if METRIC_CONFIG.get('INCLUDE_RISK_METRICS'):
        daily_drawdown = pd.read_excel(ROOT_PATH + '/data/benchmark_results/risk_metrics/'+DATASET_NAME+'/daily_drawdown.xlsx')
        final_risk_result = pd.read_excel(ROOT_PATH + '/data/benchmark_results/risk_metrics/'+DATASET_NAME+'/final_risk_result.xlsx')

        daily_drawdown = daily_drawdown.dropna(axis=1, how="any")
        final_risk_result = final_risk_result.dropna(axis=1, how="any")
        if caculate_metric_output != None:
            daily_drawdown[MODEL_NAME] = caculate_metric_output["DDD"]
            final_risk_result = final_risk_result.assign(**{MODEL_NAME: 0})
            final_risk_result.loc[0, MODEL_NAME] = caculate_metric_output["VR"]
            final_risk_result.loc[1, MODEL_NAME] = caculate_metric_output["MDD"]
        print(
            final_risk_result
        )
        plot_dataframe(daily_drawdown, PLOT_ALL_1, 'DDD', logdir)
        plot_dataframe(daily_drawdown, PLOT_ALL_2, 'DDD', logdir)
        plot_dataframe(daily_drawdown, PLOT_ALL_3, 'DDD', logdir)
        plot_dataframe(daily_drawdown, PLOT_ALL_4, 'DDD', logdir)
        plot_dataframe(daily_drawdown, PLOT_ALL_5, 'DDD', logdir)

    if METRIC_CONFIG.get('PRACTICAL_METRICS')['INCLUDE_PRACTICAL_METRICS']:
        transaction_costs_adjusted_cumulative_wealth = pd.read_excel(ROOT_PATH + '/data/benchmark_results/practical_metrics/'+DATASET_NAME+'/transaction_costs_adjusted_cumulative_wealth.xlsx')
        final_practical_result = pd.read_excel(ROOT_PATH + '/data/benchmark_results/practical_metrics/'+DATASET_NAME+'/final_practical_result.xlsx')

        transaction_costs_adjusted_cumulative_wealth = transaction_costs_adjusted_cumulative_wealth.dropna(axis=1, how="any")
        final_practical_result = final_practical_result.dropna(axis=1, how="any")
        if caculate_metric_output != None:
            transaction_costs_adjusted_cumulative_wealth[MODEL_NAME] = caculate_metric_output["TCW"]
            final_practical_result = final_practical_result.assign(**{MODEL_NAME: 0})
            final_practical_result.loc[0, MODEL_NAME] = caculate_metric_output["ATO"]
            final_practical_result.loc[1, MODEL_NAME] = caculate_metric_output["RT"]
        print(
            final_practical_result
        )
        plot_dataframe(transaction_costs_adjusted_cumulative_wealth, PLOT_ALL_1, 'TCW', logdir)
        plot_dataframe(transaction_costs_adjusted_cumulative_wealth, PLOT_ALL_2, 'TCW', logdir)
        plot_dataframe(transaction_costs_adjusted_cumulative_wealth, PLOT_ALL_3, 'TCW', logdir)
        plot_dataframe(transaction_costs_adjusted_cumulative_wealth, PLOT_ALL_4, 'TCW', logdir)
        plot_dataframe(transaction_costs_adjusted_cumulative_wealth, PLOT_ALL_5, 'TCW', logdir)



if __name__ == '__main__':
    # utils.check_update()
    load_benchmark()


import torch
from finol.config import *


def LOG_SINGLE_PERIOD_WEALTH(preds, labels):
    dot_product = torch.sum(preds * labels, dim=-1)
    log_max_dot_product = torch.log(torch.clamp(dot_product, min=1e-6))
    loss = - torch.mean(log_max_dot_product)
    return loss


def LOG_SINGLE_PERIOD_WEALTH_L2_DIVERSIFICATION(preds, labels):
    dot_product = torch.sum(preds * labels, dim=-1)
    log_max_dot_product = torch.log(torch.clamp(dot_product, min=1e-6))
    l2_norm = torch.norm(preds, p=2, dim=-1)
    loss = - torch.mean(log_max_dot_product) + LAMBDA_L2 * torch.mean(l2_norm)
    return loss


def LOG_SINGLE_PERIOD_WEALTH_L2_CONCENTRATION(preds, labels):
    dot_product = torch.sum(preds * labels, dim=-1)
    log_max_dot_product = torch.log(torch.clamp(dot_product, min=1e-6))
    l2_norm = torch.norm(preds, p=2, dim=-1)
    loss = - torch.mean(log_max_dot_product) - LAMBDA_L2 * torch.mean(l2_norm)
    return loss


def L2_DIVERSIFICATION(preds, labels):
    l2_norm = torch.norm(preds, p=2, dim=-1)
    loss = torch.mean(l2_norm)
    return loss


def L2_CONCENTRATION(preds, labels):
    l2_norm = torch.norm(preds, p=2, dim=-1)
    loss = - torch.mean(l2_norm)
    return loss


class select_criterion:
    def __init__(self):
        self.criterion_dict = {
            'LOG_SINGLE_PERIOD_WEALTH': LOG_SINGLE_PERIOD_WEALTH,
            'LOG_SINGLE_PERIOD_WEALTH_L2_DIVERSIFICATION': LOG_SINGLE_PERIOD_WEALTH_L2_DIVERSIFICATION,
            'LOG_SINGLE_PERIOD_WEALTH_L2_CONCENTRATION': LOG_SINGLE_PERIOD_WEALTH_L2_CONCENTRATION,
            'L2_DIVERSIFICATION': LOG_SINGLE_PERIOD_WEALTH_L2_CONCENTRATION,
            'L2_CONCENTRATION': LOG_SINGLE_PERIOD_WEALTH_L2_CONCENTRATION,
        }
        self.default_criterion = CRITERION_NAME

    def __call__(self, preds, labels):
        criterion = self.criterion_dict.get(
            self.default_criterion,
            None
        )

        if criterion is None:
            raise ValueError(f"Invalid criterion: {self.default_criterion}")

        return criterion(preds, labels)



import torch.nn as nn
import torch.nn.functional as F
from einops import rearrange
from finol.config import *

NUM_LAYERS = MODEL_CONFIG.get("Transformer")["NUM_LAYERS"]
NUM_HEADS = MODEL_CONFIG.get("Transformer")["NUM_HEADS"]
HIDDEN_SIZE = MODEL_CONFIG.get("Transformer")["HIDDEN_SIZE"]


class Transformer(nn.Module):
    def __init__(
            self,
            *,
            num_assets,
            num_features_augmented,
            num_features_original,
            window_size,
            **kwargs
    ):
        super().__init__()
        self.input_size = num_features_original
        self.num_feats = num_features_original
        self.output_size = num_assets
        self.num_features_augmented = num_features_augmented
        self.num_features_original = num_features_original
        self.window_size = window_size

        self.model = nn.TransformerEncoderLayer(d_model=self.input_size, nhead=NUM_HEADS, dim_feedforward=HIDDEN_SIZE, batch_first=True)
        self.dropout = nn.Dropout(p=DROPOUT)
        self.linear = nn.Linear(self.input_size, 1)

    def forward(
            self,
            x
    ):
        batch_size = x.shape[0]

        # Input Transformation
        x = x.view(batch_size, self.num_assets, self.window_size, self.num_features_original)
        x = rearrange(x, 'b m n d -> (b m) n d')

        # Temporal Representation Extraction
        output = self.model(x)
        output = output[:, -1, :]

        # Decision Making
        output = output.view(batch_size, self.num_assets, self.num_features_original)
        output = self.dropout(output)
        output = F.relu(output)
        output = self.linear(output).squeeze(-1)
        portfolio = F.softmax(output, dim=-1)
        return portfolio



import time
import torch
import os
import matplotlib

import torch.nn as nn
import matplotlib.pyplot as plt
import numpy as np
import torch.nn.functional as F
import pandas as pd
from shutil import copy2
from IPython import display
from tqdm import tqdm

from finol.data_layer.data_loader import *
from finol.model_layer.model_selector import *
from finol.optimization_layer.criterion_selector import *
from finol.optimization_layer.optimizer_selector import *
from finol.evaluation_layer.benchmark_loader import *
from finol.config import *

logdir = PARENT_PATH + '/logdir/' + str(time.strftime("%Y-%m-%d_%H-%M-%S", time.localtime()))

is_ipython = 'inline' in matplotlib.get_backend()
metadata = [
    ('TUTORIAL_MODE', TUTORIAL_MODE),
    ('TUTORIAL_NAME', TUTORIAL_NAME),
    ('Dataset', DATASET_NAME),
    ('TRAIN_END_TIMESTAMP', DATASET_SPLIT_CONFIG.get(DATASET_NAME)["TRAIN_END_TIMESTAMP"]),
    ('SCALER', SCALER),
    ('BATCH_SIZE', BATCH_SIZE[DATASET_NAME]),
    ('MODEL_NAME', MODEL_NAME),
    ('MODEL_CONFIG', MODEL_CONFIG[MODEL_NAME]),
    ('DROPOUT', DROPOUT),
    ('OPTIMIZER_NAME', OPTIMIZER_NAME),
    ('LEARNING_RATE', LEARNING_RATE),
    ('CRITERION_NAME', CRITERION_NAME),
    ('LAMBDA_L2', LAMBDA_L2),
    ('DEVICE', DEVICE),
    ('NUM_EPOCHES', NUM_EPOCHES)
]


def plot_loss_notebook(train_loss_list, val_loss_list):
    if is_ipython:
        plt.ion()
        plt.figure(figsize=(12, 5))
        plt.clf()
        plt.plot(train_loss_list, linestyle='-', marker=MARKERS[0], markevery=MARKEVERY, color='black', alpha=ALPHA, label='train loss')
        plt.plot(val_loss_list, linestyle=':', marker=MARKERS[1], markevery=MARKEVERY, color='black', alpha=ALPHA, label='val loss')
        plt.xlabel('Epoch')
        plt.ylabel('Loss')
        y_start = 1.0
        y_step = -0.05

        for i, (label, value) in enumerate(metadata):
            y_coord = y_start + i * y_step
            plt.text(1.05, y_coord, f'{label}: {value}', transform=plt.gca().transAxes, va='center', ha='left')

        plt.legend()
        plt.grid(True)
        display.clear_output(wait=True)
        display.display(plt.gcf())
        plt.tight_layout()
        plt.savefig(logdir + '/loss_' + DATASET_NAME + '.png')
        plt.clf()
        plt.close()


def plot_loss(train_loss_list, val_loss_list):
    # not_ipython = 'inline' not in matplotlib.get_backend()
    if not is_ipython:
        plt.figure(figsize=(12, 5))
        plt.plot(np.array(train_loss_list), linestyle='-', marker=MARKERS[0], markevery=MARKEVERY, color='black', alpha=ALPHA, label='train loss')
        plt.plot(np.array(val_loss_list), linestyle=':', marker=MARKERS[1], markevery=MARKEVERY, color='black', alpha=ALPHA, label='val loss')
        plt.xlabel('Epoch')
        plt.ylabel('Loss')
        plt.legend()
        plt.grid(True)
        y_start = 1.0
        y_step = -0.05

        for i, (label, value) in enumerate(metadata):
            y_coord = y_start + i * y_step
            plt.text(1.05, y_coord, f'{label}: {value}', transform=plt.gca().transAxes, va='center', ha='left')

        plt.tight_layout()
        plt.savefig(logdir + '/loss_' + DATASET_NAME + '.png')
        plt.show()


def train_model(load_dataset_output):
    torch.manual_seed(MANUAL_SEED)
    os.makedirs(logdir)
    copy2(ROOT_PATH + '/config.py', logdir)

    print(
        load_dataset_output
    )
    train_loader = load_dataset_output['train_loader']
    val_loader = load_dataset_output['val_loader']
    test_loader = load_dataset_output['test_loader']
    NUM_TRAIN_PERIODS = load_dataset_output['NUM_TRAIN_PERIODS']
    NUM_VAL_PERIODS = load_dataset_output['NUM_VAL_PERIODS']
    NUM_TEST_PERIODS = load_dataset_output['NUM_TEST_PERIODS']

    model = select_model(load_dataset_output)
    optimizer = select_optimizer(model)
    criterion = select_criterion()

    train_loss_list = []
    val_loss_list = []
    best_val_loss = float('inf')

    for e in tqdm(range(NUM_EPOCHES), desc="Training"):
        train_loss = 0
        for i, data in enumerate(train_loader, 1):
            x_data, label = data
            portfolio = model(x_data.float())
            loss = criterion(portfolio, label.float())

            optimizer.zero_grad()
            loss.backward()
            optimizer.step()
            train_loss += loss.item()

        train_loss /= len(train_loader)
        train_loss_list.append(train_loss)

        with torch.no_grad():
            val_loss = 0
            for i, data in enumerate(val_loader, 1):
                val_data, label = data
                portfolio = model(val_data.float())
                loss = criterion(portfolio, label.float())
                val_loss += loss.item()

            val_loss /= len(val_loader)
            val_loss_list.append(val_loss)

            if val_loss < best_val_loss:
                best_val_loss = val_loss
                torch.save(model, logdir + '/best_model_'+DATASET_NAME+'.pt')
            torch.save(model, logdir + '/last_model_' + DATASET_NAME + '.pt')

        if (e + 1) % 10 == 0:
            print('Epoch: {}, Train Loss: {}, Val Loss: {}'.format(e + 1, train_loss, val_loss))
            plot_loss_notebook(train_loss_list, val_loss_list)

    plot_loss(train_loss_list, val_loss_list)

    train_model_output = {
        "last_model": torch.load(logdir + '/last_model_'+DATASET_NAME+'.pt'),
        "best_model": torch.load(logdir + '/best_model_'+DATASET_NAME+'.pt'),
        "logdir": logdir,
    }
    return train_model_output


if __name__ == '__main__':
    load_dataset_output = load_dataset()
    train_model_output = train_model(load_dataset_output)